using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using System.Text;
using System.ComponentModel;
using System.Threading;
using System.Diagnostics;

/* This example will only work on Windows. As of 7/20 32 bit dotnet core Linux compatibility is limited */ 

namespace C06PKF
{
    class Program
    {

        /* NAG routine declarations copied from flcsdnet.cs*/

        [StructLayout(LayoutKind.Sequential)]/* Force sequential mapping */
        public struct Complex
        {
            public double Real_Part;
            public double Imag_Part;
        }

        [DllImport("NLW32287E_nag.dll")] /* Import from DLL, the C# compiler provides a rudimentary check of the signature */
        public static extern void C06PKF(
         ref int JOB,
         [In, Out] Complex[] X,
         [In, Out] Complex[] Y,
         ref int N,
         [Out] Complex[] WORK,
         ref int IFAIL
        );

        static void Main(string[] args)
        {
            int ifail;
            int n = 5000000;
            Complex[] x = new Complex[n];
            Complex[] y = new Complex[n];
            Complex[] work = new Complex[4 * n];

            for (int j = 0; j < n; j++)
            {
                x[j].Real_Part = (double)1.0 + Math.Cos(j);
                x[j].Imag_Part = (double)-0.5 + Math.Sin(j);
            }


            for (int i = 0; i < n; i++)
            {
                y[i].Real_Part = (double)0.5 + Math.Sin(i);
                y[i].Imag_Part = (double)-0.25 + Math.Cos(i);
            }

            ifail = 1;
            int job = 2;

            Stopwatch stopWatch = new Stopwatch();
            stopWatch.Start();

            C06PKF(ref job, x, y, ref n, work, ref ifail);

            stopWatch.Stop();

            TimeSpan ts = stopWatch.Elapsed;

            if (ifail != 0)
            {
                Console.WriteLine("An error occurred: ifail = {0}\n", ifail);
            }
            else
            {
                Console.WriteLine("For n = {0}, duration = {1} sec\n", n, ts.TotalSeconds);
            }
        }
    }
}
